import os
import json
from unittest.mock import Mock, patch

from ligo.gracedb import rest
import pkg_resources
import pytest

from ..tasks import orchestrator
from . import test_tasks_skymaps
from . import resource_json


@pytest.mark.parametrize(
    'group,pipeline', [['CBC', 'gstlal'], ['Burst', 'CWB'], ['Burst', 'LIB']])
def test_handle_superevent(monkeypatch, group, pipeline):
    """Test a superevent is dispatched to the correct annotation task based on
    its preferred event's search group."""
    alert = {
        'alert_type': 'new',
        'object': {
            'superevent_id': 'S1234',
            't_start': 1214714160,
            't_end': 1214714164,
            'preferred_event': 'G1234'
        }
    }

    def get_superevent(superevent_id):
        assert superevent_id == 'S1234'
        return {'preferred_event': 'G1234'}

    def get_event(graceid):
        assert graceid == 'G1234'
        return {'group': group, 'pipeline': pipeline,
                'instruments': 'H1,L1,V1', 'graceid': 'G1234'}

    def download(filename, graceid):
        if '.fits' in filename:
            return test_tasks_skymaps.toy_3d_fits_filecontents()
        elif filename == 'source_classification.json' and group == 'CBC':
            return json.dumps({'Prob NS2': 0, 'Prob EMbright': 0})
        elif filename == 'psd.xml.gz':
            return pkg_resources.resource_filename(__name__, 'data/psd.xml.gz')
        elif filename == 'S1234-1-Preliminary.xml':
            return b'fake VOEvent file contents'
        else:
            raise ValueError

    def create_voevent(*args, **kwargs):
        return 'S1234-1-Preliminary.xml'

    create_circular = Mock()
    plot_volume = Mock()
    plot_allsky = Mock()
    send = Mock()

    monkeypatch.setattr('gwcelery.tasks.gcn.send.run', send)
    monkeypatch.setattr('gwcelery.tasks.skymaps.plot_allsky.run', plot_allsky)
    monkeypatch.setattr('gwcelery.tasks.skymaps.plot_volume.run', plot_volume)
    monkeypatch.setattr('gwcelery.tasks.gracedb.download.run', download)
    monkeypatch.setattr('gwcelery.tasks.gracedb.get_event.run', get_event)
    monkeypatch.setattr('gwcelery.tasks.gracedb.create_voevent.run',
                        create_voevent)
    monkeypatch.setattr('gwcelery.tasks.gracedb.get_superevent.run',
                        get_superevent)
    monkeypatch.setattr('gwcelery.tasks.circulars.create_circular.run',
                        create_circular)

    # Run function under test
    orchestrator.handle_superevent(alert)

    plot_allsky.assert_called_once()
    plot_volume.assert_called_once()
    send.assert_called_once()
    create_circular.assert_called_once()


def mock_download(filename, graceid, *args, **kwargs):
    assert graceid == 'T250822'
    filenames = {'coinc.xml': 'coinc.xml',
                 'psd.xml.gz': 'psd.xml.gz',
                 'ranking_data.xml.gz': 'ranking_data_G322589.xml.gz'}
    return pkg_resources.resource_string(
        __name__, os.path.join('data', filenames[filename]))


@patch(
    'gwcelery.tasks.gracedb.get_event.run',
    return_value={'graceid': 'T250822', 'group': 'CBC', 'pipeline': 'gstlal'})
@patch('gwcelery.tasks.gracedb.download.run', mock_download)
@patch('gwcelery.tasks.em_bright.classifier.run')
@patch('gwcelery.tasks.bayestar.localize.run')
@patch('gwcelery.tasks.gracedb.client', autospec=rest.GraceDb)
def test_handle_cbc_event(mock_gracedb, mock_localize, mock_classifier,
                          mock_get_event):
    """Test that an LVAlert message for a newly uploaded PSD file triggers
    BAYESTAR."""
    alert = resource_json(__name__, 'data/lvalert_psd.json')
    orchestrator.handle_cbc_event(alert)
    mock_localize.assert_called_once()
    mock_classifier.assert_called_once()


@patch('gwcelery.tasks.gracedb.download.run', mock_download)
@patch('gwcelery.tasks.em_bright.classifier.run')
@patch('gwcelery.tasks.bayestar.localize.run')
@patch('gwcelery.tasks.gracedb.client', autospec=rest.GraceDb)
def test_handle_cbc_event_ignored(mock_gracedb, mock_localize,
                                  mock_classifier):
    """Test that unrelated LVAlert messages do not trigger BAYESTAR."""
    alert = resource_json(__name__, 'data/lvalert_detchar.json')
    orchestrator.handle_cbc_event(alert)
    mock_localize.assert_not_called()
    mock_classifier.assert_not_called()


@patch('gwcelery.tasks.raven.coincidence_search')
def test_handle_superevent_creation(mock_raven_coincidence_search):
    """Test dispatch of an LVAlert message for a superevent creation."""
    # Test LVAlert payload.
    alert = resource_json(__name__, 'data/lvalert_superevent_creation.json')

    # Run function under test
    orchestrator.handle_superevents_externaltriggers(alert)

    # Check that the correct tasks were dispatched.
    mock_raven_coincidence_search.assert_called_once_with('S180616h',
                                                          alert['object'])


@patch('gwcelery.tasks.ligo_fermi_skymaps.create_combined_skymap')
def test_handle_superevent_emcoinc_label(mock_create_combined_skymap):
    """Test dispatch of an LVAlert message for a superevent EM_COINC label
    application."""
    alert = resource_json(__name__, 'data/lvalert_superevent_label.json')
    orchestrator.handle_superevents_externaltriggers(alert)
    mock_create_combined_skymap.assert_called_once_with('S180616h')



import pytest
from unittest.mock import patch

from ..tasks import gracedb, superevents
from . import resource_json


def test_update_preferred_event(monkeypatch):
    class T0212TTPResponse(object):
        def json(self):
            return resource_json(__name__, 'data/T0212_S0039_preferred.json')

    class NewEventTTPResponse(object):
        def json(self):
            return dict(graceid="T1234",
                        instruments="I1,J1,K1,L1,M1",
                        group="CBC",
                        superevent="some_superevent",
                        far=1e-30,
                        snr=30.0)

    class FakeDb(object):
        def __init__(self):
            self.service_url = 'service_url'

        def event(self, gid):
                return T0212TTPResponse()

        def updateSuperevent(self, *args, **kwargs):    # noqa: N802
            pass

    g = FakeDb()
    monkeypatch.setattr('gwcelery.tasks.gracedb.client', g)
    # call update _update_superevent such that updateSuperevent is called
    with patch.object(g, 'updateSuperevent') as p:
        superevents._update_superevent('S0039',
                                       'T0212',
                                       NewEventTTPResponse().json(),
                                       None,
                                       None)
        p.assert_called_with('S0039', preferred_event='T1234',
                             t_start=None, t_end=None, t_0=None)


def test_get_superevents(monkeypatch):
    class FakeDb(object):
        def __init__(self):
            self.service_url = 'service_url'

        def superevents(self, **kwargs):
            response = resource_json(__name__, 'data/superevents.json')
            return (s for s in response['superevents'])

    monkeypatch.setattr('gwcelery.tasks.gracedb.client', FakeDb())

    gid_1 = 'T0212'  # present in fake superevent queue
    gid_2 = 'T0219'  # present but not preferred event
    gid_3 = 'T9999'  # absent in fake superevent queue
    superevent_id_1, preferred_flag_1, r_1 = gracedb.get_superevents(gid_1)
    superevent_id_2, preferred_flag_2, r_2 = gracedb.get_superevents(gid_2)
    superevent_id_3, preferred_flag_3, r_3 = gracedb.get_superevents(gid_3)

    assert superevent_id_1 == 'S0039' and preferred_flag_1
    assert superevent_id_2 == 'S0041' and not preferred_flag_2
    assert superevent_id_3 is None


def test_parse_trigger_raising():
    garbage_payload = dict(some_value=999, something_else=99)
    with pytest.raises(KeyError):
        superevents.handle(garbage_payload)


# FIXME lot of redundancy, implement setUp tearDown for below
def test_parse_trigger_cbc_1(monkeypatch):
    """New trigger G000000, less significant than already
    existing superevent. Superevent window much larger than event
    window of G000000"""
    class T0212TTPResponse(object):
        def json(self):
            return resource_json(__name__, 'data/T0212_S0039_preferred.json')

    class FakeDb(object):
        def __init__(self):
            self.service_url = 'service_url'

        def superevents(self, **kwargs):
            response = resource_json(__name__, 'data/superevents.json')
            return (s for s in response['superevents'])

        def updateSuperevent(self, *args, **kwargs):    # noqa: N802
            pass

        def addEventToSuperevent(self, *args, **kwargs):    # noqa: N802
            pass

        def event(self, gid):
            if gid == "T0212":
                return T0212TTPResponse()
            else:
                raise ValueError("Called with incorrect preferred event")
    g = FakeDb()
    monkeypatch.setattr('gwcelery.tasks.gracedb.client', g)
    payload = \
        resource_json(__name__, 'data/mock_trigger_new_G000000_cbc.json')
    # New trigger G000000 time falls in S0039 window
    # addEventToSuperevent should be called
    # updateSuperevent should not be updated
    with patch.object(g, 'addEventToSuperevent') as p1, \
            patch.object(g, 'updateSuperevent') as p2:
        superevents.handle(payload)
        p1.assert_called_once()
        p2.assert_not_called()


def test_parse_trigger_cbc_2(monkeypatch):
    """New trigger G000003, more significant than already
    existing superevent. Superevent window is much larger that
    event window of G000000"""
    class T0212TTPResponse(object):
        def json(self):
            return resource_json(__name__, 'data/T0212_S0039_preferred.json')

    class FakeDb(object):
        def __init__(self):
            self.service_url = 'service_url'

        def superevents(self, **kwargs):
            response = resource_json(__name__, 'data/superevents.json')
            return (s for s in response['superevents'])

        def addEventToSuperevent(self, *args, **kwargs):    # noqa: N802
            pass

        def updateSuperevent(self, *args, **kwargs):    # noqa: N802
            pass

        def event(self, gid):
            if gid == "T0212":
                return T0212TTPResponse()
            else:
                raise ValueError("Called with incorrect preferred event")
    g = FakeDb()
    monkeypatch.setattr('gwcelery.tasks.gracedb.client', g)
    # New trigger G000000 falls in S0039 window
    payload = \
        resource_json(__name__, 'data/mock_trigger_new_G000003_cbc.json')
    # addEventToSuperevent should be called
    # preferred event should be updated
    with patch.object(g, 'addEventToSuperevent') as p1, \
            patch.object(g, 'updateSuperevent') as p2:
        superevents.handle(payload)
        p1.assert_called_once()
        p2.assert_called_once_with('S0039', preferred_event='G000003',
                                   t_0=None, t_start=None, t_end=None)


def test_parse_trigger_cbc_3(monkeypatch):
    """New trigger G000001, not present among superevents
    New superevent created"""
    class FakeDb(object):
        def __init__(self):
            self.service_url = 'service_url'

        def createSuperevent(self, *args, **kwargs):    # noqa: N802
            pass

        def superevents(self, **kwargs):
            response = resource_json(__name__, 'data/superevents.json')
            return (s for s in response['superevents'])

    g = FakeDb()
    monkeypatch.setattr('gwcelery.tasks.gracedb.client', g)
    payload = \
        resource_json(__name__, 'data/mock_trigger_new_G000001_cbc.json')
    # G000001 absent in any superevent window, new superevent created
    # createSuperevent should be called exactly once
    with patch.object(g, 'createSuperevent') as p:
        superevents.handle(payload)
        p.assert_called_once()


def test_parse_trigger_cbc_4(monkeypatch):
    """New trigger G000002, doesn't pass far threshold"""
    class FakeDb(object):
        def __init__(self):
            self.service_url = 'service_url'

        def superevents(self, **kwargs):
            response = resource_json(__name__, 'data/superevents.json')
            return (s for s in response['superevents'])

        def updateSuperevent(self, *args, **kwargs):    # noqa: N802
            pass

        def addEventToSuperevent(self, *args, **kwargs):    # noqa: N802
            pass

        def createSuperevent(self, *args, **kwargs):    # noqa: N802
            pass

    g = FakeDb()
    monkeypatch.setattr('gwcelery.tasks.gracedb.client', g)
    payload = \
        resource_json(__name__, 'data/mock_trigger_new_G000002_cbc.json')
    superevents.handle(payload)
    # neither method is called due to low far
    with patch.object(g, 'addEventToSuperevent') as p1, \
            patch.object(g, 'createSuperevent') as p2:
        superevents.handle(payload)
        p1.assert_not_called()
        p2.assert_not_called()


def test_parse_trigger_burst_1(monkeypatch):
    """New cwb trigger G000005 with gpstime lying partially in
    S0039 window, not more significant than existing preferred
    event, superevent window changed.
    """
    class T0212TTPResponse(object):
        def json(self):
            return resource_json(__name__, 'data/T0212_S0039_preferred.json')

    class FakeDb(object):
        def __init__(self):
            self.service_url = 'service_url'

        def superevents(self, **kwargs):
            response = resource_json(__name__, 'data/superevents.json')
            return (s for s in response['superevents'])

        def updateSuperevent(self, *args, **kwargs):    # noqa: N802
            pass

        def addEventToSuperevent(self, *args, **kwargs):    # noqa: N802
            pass

        def event(self, gid):
            if gid == "T0212":
                return T0212TTPResponse()
            else:
                raise ValueError("Called with incorrect preferred event")
    g = FakeDb()
    monkeypatch.setattr('gwcelery.tasks.gracedb.client', g)
    payload = \
        resource_json(__name__, 'data/mock_trigger_new_G000005_burst.json')
    # preferred event should not be updated
    # superevent window should change
    with patch.object(g, 'addEventToSuperevent') as p1, \
            patch.object(g, 'updateSuperevent') as p2:
        superevents.handle(payload)
        p1.assert_called_once()
        p2.assert_called_once_with('S0039', preferred_event=None, t_0=None,
                                   t_end=pytest.approx(1163905239, abs=1),
                                   t_start=pytest.approx(1163905214, abs=1))


def test_parse_trigger_burst_2(monkeypatch):
    """New LIB trigger G000006 with gpstime lying partially in
    S0039 window, more significant than already existing preferred
    event. superevent window changed
    """
    class T0212TTPResponse(object):
        def json(self):
            return resource_json(__name__, 'data/T0212_S0039_preferred.json')

    class FakeDb(object):
        def __init__(self):
            self.service_url = 'service_url'

        def superevents(self, **kwargs):
            response = resource_json(__name__, 'data/superevents.json')
            return (s for s in response['superevents'])

        def updateSuperevent(self, *args, **kwargs):    # noqa: N802
            pass

        def addEventToSuperevent(self, *args, **kwargs):    # noqa: N802
            pass

        def event(self, gid):
            if gid == "T0212":
                return T0212TTPResponse()
            else:
                raise ValueError("Called with incorrect preferred event")
    g = FakeDb()
    monkeypatch.setattr('gwcelery.tasks.gracedb.client', g)
    payload = \
        resource_json(__name__, 'data/mock_trigger_new_G000006_burst.json')
    # preferred event not be updated, in spite of lower far of new event
    # because existing preferred event is CBC
    with patch.object(g, 'addEventToSuperevent') as p1, \
            patch.object(g, 'updateSuperevent') as p2:
        superevents.handle(payload)
        p1.assert_called_once()
        p2.assert_called_once_with('S0039', t_0=None, preferred_event=None,
                                   t_end=pytest.approx(1163905239, abs=1),
                                   t_start=pytest.approx(1163905214, abs=1))


def test_parse_trigger_burst_3(monkeypatch):
    """New LIB trigger G000007, not present among superevents
    New superevent created.
    Q_mean = frequency_mean = 100., hence d_t_start = d_t_end = 1s
    """
    class FakeDb(object):
        def __init__(self):
            self.service_url = 'service_url'

        def createSuperevent(self, *args, **kwargs):    # noqa: N802
            pass

        def superevents(self, **kwargs):
            response = resource_json(__name__, 'data/superevents.json')
            return (s for s in response['superevents'])

    g = FakeDb()
    monkeypatch.setattr('gwcelery.tasks.gracedb.client', g)
    payload = \
        resource_json(__name__, 'data/mock_trigger_new_G000007_burst.json')
    # G000007 absent in any superevent window, new superevent created
    with patch.object(g, 'createSuperevent') as p:
        superevents.handle(payload)
        p.assert_called_once_with(pytest.approx(1163905248.5),
                                  pytest.approx(1163905249.5),
                                  pytest.approx(1163905250.5),
                                  preferred_event='G000007')


def test_parse_trigger_burst_4(monkeypatch):
    """New CWB trigger G000008, not present among superevents
    New superevent created.
    extra attribute duration = 0.02s
    """
    class FakeDb(object):
        def __init__(self):
            self.service_url = 'service_url'

        def createSuperevent(self, *args, **kwargs):    # noqa: N802
            self._create_superevent_called += 1
            # values are such that the d_start and d_t_end are 1s
            # assert t_end - t_start == 0.04
            assert args[2] - args[0] == pytest.approx(0.04)
            assert kwargs.get('preferred_event') == 'G000008'

        def superevents(self, **kwargs):
            response = resource_json(__name__, 'data/superevents.json')
            return (s for s in response['superevents'])

    g = FakeDb()
    monkeypatch.setattr('gwcelery.tasks.gracedb.client', g)
    payload = \
        resource_json(__name__, 'data/mock_trigger_new_G000008_burst.json')
    # G000008 absent in any superevent window, new superevent created
    with patch.object(g, 'createSuperevent') as p:
        superevents.handle(payload)
        p.assert_called_once()



"""Subsystem for sending, receiving, and processing Gamma-ray Coordinates
Network [GCN]_ notices.

References
----------

.. [GCN] https://gcn.gsfc.nasa.gov
"""
import socket
import struct
import time

from celery.utils.log import get_task_logger
from celery_eternal import EternalTask, EternalProcessTask
from gcn.voeventclient import _get_now_iso8601, _recv_packet, _send_packet
from gcn import get_notice_type, NoticeType
import gcn

from ...celery import app
from ..core import DispatchHandler

log = get_task_logger(__name__)


IAMALIVE = '''<?xml version="1.0" encoding="UTF-8"?>
<trn:Transport role="iamalive" version="1.0"
xmlns:trn="http://telescope-networks.org/schema/Transport/v1.1"
xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
xsi:schemaLocation="http://telescope-networks.org/schema/Transport/v1.1
http://telescope-networks.org/schema/Transport-v1.1.xsd">
<Origin>{}</Origin><TimeStamp>{}</TimeStamp>
</trn:Transport>'''


KEEPALIVE_TIME = 60  # Time in seconds between keepalives


def _host_port(address):
    host, port = address.split(':')
    return host, int(port)


@app.task(base=EternalTask, bind=True)
def broker(self):
    """Single-client VOEvent broker for sending notices to GCN.

    This is a basic VOEvent broker. It binds to the address
    :obj:`~gwcelery.celery.Base.gcn_broker_address` and accepts
    one connection at a time from any host whose address is listed in
    :obj:`~gwcelery.celery.Base.gcn_broker_accept_addresses`.
    """
    fqdn = socket.getfqdn()
    host, port = _host_port(app.conf['gcn_broker_address'])
    accept_hosts = [socket.gethostbyname(host) for host in
                    app.conf['gcn_broker_accept_addresses']]

    with socket.socket() as sock:
        sock.settimeout(1.0)
        sock.bind((host, port))
        log.info('bound to %s', app.conf['gcn_broker_address'])
        sock.listen(0)
        while not self.is_aborted():
            try:
                conn, (addr, _) = sock.accept()
            except socket.timeout:
                continue
            if addr in accept_hosts:
                log.info('accepted connection from remote host %s', addr)
                break
            else:
                log.error('denied connection from remote host %s', addr)
                conn.close()
        else:  # self.is_aborted()
            return

    with conn:
        conn.settimeout(1.0)
        conn.setsockopt(socket.SOL_SOCKET, socket.SO_LINGER,
                        struct.pack('ii', 1, 0))
        conn.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, 1)

        last_sent = float('inf')

        while not self.is_aborted():
            # Get next payload from queue in first-in, first-out fashion
            payload = self.backend.client.lindex(_queue_name, 0)
            now = time.monotonic()
            if payload is not None:
                log.info('sending payload of %d bytes', len(payload))
                _send_packet(conn, payload)
                last_sent = now
                self.backend.client.lpop(_queue_name)
            elif now - last_sent > KEEPALIVE_TIME:
                timestamp = _get_now_iso8601()
                payload = IAMALIVE.format(fqdn, timestamp).encode('utf-8')
                log.info('sending keepalive')
                _send_packet(conn, payload)
                last_sent = now
            else:
                # Read (but ignore) any inbound packet
                try:
                    _recv_packet(conn)
                except socket.timeout:
                    pass


_queue_name = broker.name + '.voevent-queue'


@app.task(bind=True, ignore_result=True, shared=False)
def send(self, payload):
    """Send a VOEvent to GCN.

    Under the hood, this task just pushes the payload onto a Redis queue,
    and :func:`~gwcelery.tasks.gcn.broker` sends it."""
    self.backend.client.rpush(_queue_name, payload)


class _VOEventDispatchHandler(DispatchHandler):

    def process_args(self, payload, root):
        notice_type = get_notice_type(root)

        # Just cast to enum for prettier log messages
        try:
            notice_type = NoticeType(notice_type)
        except ValueError:
            pass

        return notice_type, (payload,), {}


handler = _VOEventDispatchHandler()
"""Function decorator to register a handler callback for specified GCN notice
types. The decorated function is turned into a Celery task, which will be
automatically called whenever a matching GCN notice is received.

Parameters
----------
\*keys
    List of GCN notice types to accept
\*\*kwargs
    Additional keyword arguments for :meth:`celery.Celery.task`.

Examples
--------
Declare a new handler like this::

    @gcn.handler(gcn.NoticeType.FERMI_GBM_GND_POS,
                 gcn.NoticeType.FERMI_GBM_FIN_POS)
    def handle_fermi(payload):
        root = lxml.etree.fromstring(payload)
        # do work here...
"""


@app.task(base=EternalProcessTask, shared=False)
def listen():
    """Listen to GCN notices forever. GCN notices are dispatched asynchronously
    to tasks that have been registered with
    :meth:`gwcelery.tasks.gcn.handler`."""
    host, port = _host_port(app.conf['gcn_client_address'])
    gcn.listen(host, port, handler=handler.dispatch)



from unittest.mock import patch

from ligo.gracedb import rest
import pkg_resources

from . import resource_json
from . import test_tasks_skymaps
from ..tasks import ligo_fermi_skymaps


def mock_get_superevent(graceid):
    return resource_json(__name__, 'data/mock_superevent_object.json')


def mock_get_log(graceid):
    if graceid == 'S12345':
        return resource_json(__name__, 'data/gracedb_setrigger_log.json')
    elif graceid == 'E12345':
        return resource_json(__name__, 'data/gracedb_externaltrigger_log.json')
    else:
        raise ValueError


def mock_download(filename, graceid):
    """Mocks GraceDb download functionality"""
    if graceid == 'S12345' and filename == 'bayestar.fits.gz':
        return test_tasks_skymaps.toy_3d_fits_filecontents()
    elif (graceid == 'E12345' and
          filename == ('nasa.gsfc.gcn_Fermi%23GBM_Gnd_Pos_2017-08-17'
                       + 'T12%3A41%3A06.47_524666471_57-431.xml')):
        return pkg_resources.resource_string(
                   __name__, 'data/externaltrigger_original_data.xml'
               )
    else:
        raise ValueError


def mock_get_file_contents(heasarc_link):
    """Mocks astropy get_file_contents functionality"""
    assert heasarc_link == (
        'http://heasarc.gsfc.nasa.gov/FTP/fermi/data/gbm/'
        + 'triggers/2017/bn170817529/current/'
    )
    return test_tasks_skymaps.toy_fits_filecontents()


@patch('gwcelery.tasks.gracedb.get_superevent', mock_get_superevent)
@patch('gwcelery.tasks.gracedb.get_log', mock_get_log)
@patch('gwcelery.tasks.gracedb.download', mock_download)
@patch('astropy.utils.data.get_file_contents', mock_get_file_contents)
@patch('gwcelery.tasks.gracedb.client', autospec=rest.GraceDb)
def test_create_combined_skymap(graceid):
    """Test creating combined LVC and Fermi skymap"""
    # Run function under test
    ligo_fermi_skymaps.create_combined_skymap('S12345')



"""Module containing the computation of p_astro by source category
   See https://dcc.ligo.org/LIGO-T1800072 for details.
"""
import io
import json
import os
import sqlite3

from celery.utils.log import get_task_logger
from celery.exceptions import Ignore
from glue.ligolw import ligolw
from glue.ligolw.ligolw import LIGOLWContentHandler
from glue.ligolw import array as ligolw_array
from glue.ligolw import param as ligolw_param
from glue.ligolw import dbtables
from glue.ligolw import utils as ligolw_utils
from glue.ligolw import lsctables
from lalinspiral import thinca
from lal import rate
from ligo.p_astro import SourceType, MarginalizedPosterior
import numpy as np
import pkg_resources

from ..celery import app
from ..util import PromiseProxy

log = get_task_logger(__name__)

# adapted from gstlal far.py RankingStatPDF


class RankingStatPDF(object):
    ligo_lw_name_suffix = "gstlal_inspiral_rankingstatpdf"

    @classmethod
    def from_xml(cls, xml, name):
        """
        Find the root of the XML tree containing the
        serialization of this object
        """
        xml, = [elem for elem in
                xml.getElementsByTagName(ligolw.LIGO_LW.tagName)
                if elem.hasAttribute("Name") and
                elem.Name == "%s:%s" % (name, cls.ligo_lw_name_suffix)]
        # create a uninitialized instance
        self = super().__new__(cls)
        # populate from XML
        self.noise_lr_lnpdf = rate.BinnedLnPDF.from_xml(xml, "noise_lr_lnpdf")
        self.signal_lr_lnpdf = rate.BinnedLnPDF.from_xml(xml,
                                                         "signal_lr_lnpdf")
        self.zero_lag_lr_lnpdf = rate.BinnedLnPDF.from_xml(
            xml, "zero_lag_lr_lnpdf")
        return self


def parse_likelihood_control_doc(xmldoc):
    name = "gstlal_inspiral_likelihood"
    rankingstatpdf = RankingStatPDF.from_xml(xmldoc, name)
    if rankingstatpdf is None:
        raise ValueError("document does not contain likelihood ratio data")
    return rankingstatpdf


@ligolw_array.use_in
@ligolw_param.use_in
@lsctables.use_in
class ContentHandler(LIGOLWContentHandler):
    pass


def get_ln_f_over_b(ranking_data_bytes, ln_likelihood_ratios):
    ranking_data_xmldoc, _ = ligolw_utils.load_fileobj(
        io.BytesIO(ranking_data_bytes), contenthandler=ContentHandler)
    rankingstatpdf = parse_likelihood_control_doc(ranking_data_xmldoc)
    # affect the zeroing of the PDFs below threshold by hacking the
    # histograms. Do the indexing ourselves to not 0 the bin @ threshold
    ln_likelihood_ratio_threshold = \
        app.conf['p_astro_gstlal_ln_likelihood_threshold']
    noise_lr_lnpdf = rankingstatpdf.noise_lr_lnpdf
    rankingstatpdf.noise_lr_lnpdf.array[
        :noise_lr_lnpdf.bins[0][ln_likelihood_ratio_threshold]] \
        = 0.
    rankingstatpdf.noise_lr_lnpdf.normalize()
    signal_lr_lnpdf = rankingstatpdf.signal_lr_lnpdf
    rankingstatpdf.signal_lr_lnpdf.array[
        :signal_lr_lnpdf.bins[0][ln_likelihood_ratio_threshold]] \
        = 0.
    rankingstatpdf.signal_lr_lnpdf.normalize()
    zero_lag_lr_lnpdf = rankingstatpdf.zero_lag_lr_lnpdf
    rankingstatpdf.zero_lag_lr_lnpdf.array[
        :zero_lag_lr_lnpdf.bins[0][ln_likelihood_ratio_threshold]] \
        = 0.
    rankingstatpdf.zero_lag_lr_lnpdf.normalize()

    f = rankingstatpdf.signal_lr_lnpdf
    b = rankingstatpdf.noise_lr_lnpdf
    ln_f_over_b = \
        np.array([f[ln_lr, ] - b[ln_lr, ] for ln_lr in ln_likelihood_ratios])
    if np.isnan(ln_f_over_b).any():
        raise ValueError("NaN encountered in ranking statistic PDF ratios")
    if np.isinf(np.exp(ln_f_over_b)).any():
        raise ValueError(
            "infinity encountered in ranking statistic PDF ratios")
    return ln_f_over_b


def get_event_ln_likelihood_ratio_svd_endtime_mass(coinc_bytes):
    coinc_xmldoc, _ = ligolw_utils.load_fileobj(
        io.BytesIO(coinc_bytes), contenthandler=ContentHandler)
    coinc_event, = lsctables.CoincTable.get_table(coinc_xmldoc)
    coinc_inspiral, = lsctables.CoincInspiralTable.get_table(coinc_xmldoc)
    sngl_inspiral = lsctables.SnglInspiralTable.get_table(coinc_xmldoc)

    assert all([sngl_inspiral[i].Gamma0 == sngl_inspiral[i+1].Gamma0
                for i in range(len(sngl_inspiral)-1)]), \
        "svd bank different between ifos!"
    return (coinc_event.likelihood,
            int(sngl_inspiral[0].Gamma1),
            coinc_inspiral.end_time,
            coinc_inspiral.mass)


def load_search_database():
    filename = app.conf['p_astro_gstlal_trigger_db']
    # FIXME the gstlal trigger database path is specific to the CIT
    # cluster. Gentle exit against opening a non-existant database outside CIT
    if not os.path.exists(filename):
        raise Ignore(
            "Gstlal trigger database {} not found".format(filename))
    return sqlite3.connect('file:{}?mode=ro'.format(filename), uri=True)


# Lazily load search database
connection = PromiseProxy(load_search_database)


def load_search_results(end_time, mass, ln_likelihood_ratio_threshold):
    """Queries SQLite file for background trigger data.
    The query has an extra check to make sure the event is not
    double counted.
    """
    cur = connection.cursor()

    xmldoc = dbtables.get_xml(connection)
    definer_id = lsctables.CoincDefTable.get_table(xmldoc).get_coinc_def_id(
        thinca.InspiralCoincDef.search,
        thinca.InspiralCoincDef.search_coinc_type,
        create_new=False)
    cur.execute("""
SELECT
        coinc_event.likelihood,
        (
        SELECT sngl_inspiral.Gamma1
        FROM
                sngl_inspiral JOIN coinc_event_map ON (
                        coinc_event_map.table_name == "sngl_inspiral"
                        AND coinc_event_map.event_id == sngl_inspiral.event_id
                )
        WHERE
        coinc_event_map.coinc_event_id == coinc_inspiral.coinc_event_id
        LIMIT 1
        ),
        EXISTS (
    SELECT
*
    FROM
time_slide
    WHERE
time_slide.time_slide_id == coinc_event.time_slide_id
AND time_slide.offset != 0
        )
FROM
        coinc_event JOIN coinc_inspiral ON (
    coinc_inspiral.coinc_event_id == coinc_event.coinc_event_id
        )
WHERE
        coinc_event.coinc_def_id == ?
        AND coinc_event.likelihood >= ?
        AND coinc_inspiral.end_time != ?
        AND coinc_inspiral.mass != ?""", (definer_id,
                                          ln_likelihood_ratio_threshold,
                                          end_time,
                                          mass))
    ln_likelihood_ratio, svd_banks, is_background = np.array(cur.fetchall()).T
    background_ln_likelihood_ratios = \
        ln_likelihood_ratio[is_background.astype(bool)]
    zerolag_ln_likelihood_ratios = \
        ln_likelihood_ratio[np.logical_not(is_background.astype(bool))]
    svd_banks = svd_banks.astype(int)

    return background_ln_likelihood_ratios, \
        zerolag_ln_likelihood_ratios, svd_banks


def load_counts(name):
    # FIXME txt files will need to be queried from sqlite dbs
    filename = pkg_resources.resource_filename(
        __name__, '../data/p_astro_gstlal/{}_wellfound_hits.txt'.format(name))

    # Construction of the weights
    a = np.recfromtxt(filename, names=True)['hit_count']
    a_hat = a / a.sum()
    return a_hat


# Lazily load weights
a_hat_bns = PromiseProxy(load_counts, ('bns',))
a_hat_nsbh = PromiseProxy(load_counts, ('nsbh',))
a_hat_bbh = PromiseProxy(load_counts, ('bbh',))


def get_counts_instance(ln_f_over_b,
                        svd_bank_nums,
                        prior_type="Uniform"):
    num_svd_bins = len(a_hat_bns)

    w_bns = num_svd_bins*np.take(a_hat_bns, svd_bank_nums)
    w_nsbh = num_svd_bins*np.take(a_hat_nsbh, svd_bank_nums)
    w_bbh = num_svd_bins*np.take(a_hat_bbh, svd_bank_nums)

    num_f_over_b = len(ln_f_over_b)
    w_terr = np.ones(num_f_over_b)

    fb = np.exp(ln_f_over_b)

    return MarginalizedPosterior(f_divby_b=fb, prior_type=prior_type,
                                 terr_source=SourceType(label="Terr",
                                                        w_fgmc=w_terr),
                                 fix_sources={"Terr": num_f_over_b},
                                 bns_inst=SourceType(label="BNS",
                                                     w_fgmc=w_bns),
                                 bbh_inst=SourceType(label="BBH",
                                                     w_fgmc=w_bbh),
                                 nsbh_inst=SourceType(label="NSBH",
                                                      w_fgmc=w_nsbh),
                                 verbose=False)


@app.task(shared=False)
def compute_p_astro(files):
    """
    Task to compute `p_astro` by source category.
    """
    coinc_bytes, ranking_data_bytes = files

    log.info(
        'Fetching ln_likelihood_ratio, svd bin, endtime, mass from coinc.xml')
    event_ln_likelihood_ratio, event_svd, event_endtime, event_mass = \
        get_event_ln_likelihood_ratio_svd_endtime_mass(coinc_bytes)

    ln_likelihood_ratio_threshold = \
        app.conf['p_astro_gstlal_ln_likelihood_threshold']

    log.info('Querying trigger db to fetch zerolag ln_likelihood_ratios')

    background_ln_likelihood_ratios, zerolag_ln_likelihood_ratios, \
        svd_banks = load_search_results(event_endtime,
                                        event_mass,
                                        ln_likelihood_ratio_threshold)

    svd_banks = np.append(svd_banks, event_svd)
    zerolag_ln_likelihood_ratios = np.append(
        zerolag_ln_likelihood_ratios, event_ln_likelihood_ratio)

    log.info('Computing f_over_b from ranking_data.xml.gz')
    ln_f_over_b = get_ln_f_over_b(ranking_data_bytes,
                                  zerolag_ln_likelihood_ratios)

    log.info('Creating FGMC MarginalizedPosterior')
    event_counts_instance = \
        get_counts_instance(ln_f_over_b,
                            svd_banks,
                            prior_type=app.conf['p_astro_gstlal_prior_type'])

    num_f_over_b = len(ln_f_over_b)

    categories = ['BNS', 'BBH', 'NSBH', 'Terr']
    p_astro_values = {}
    for category in categories:
        p_astro_values[category] = \
            event_counts_instance.pastro(categories=[category],
                                         trigger_idx=num_f_over_b-1)[0]
    return json.dumps(p_astro_values)



import json

import numpy as np
import pkg_resources
import pytest

from .. import app
from ..tasks import p_astro_gstlal


@pytest.fixture
def ranking_data_bytes():
    return pkg_resources.resource_string(
        __name__, 'data/ranking_data_G322589.xml.gz')


@pytest.fixture
def coinc_bytes():
    return pkg_resources.resource_string(
        __name__, 'data/coinc_G322589.xml')


@pytest.fixture(autouse=True, scope='module')
def mock_trigger_db():
    old = app.conf['p_astro_gstlal_trigger_db']
    app.conf['p_astro_gstlal_trigger_db'] = pkg_resources.resource_filename(
        __name__, 'data/p_astro_gstlal_trigger_db.sqlite')
    yield
    app.conf['p_astro_gstlal_trigger_db'] = old


def test_get_ln_f_over_b(ranking_data_bytes):
    """Test to check :meth:`get_ln_f_over_b` returns values
    which are not inf or nan for high ln_likelihood_ratio"""
    ln_f_over_b = p_astro_gstlal.get_ln_f_over_b(
        ranking_data_bytes, [100., 200., 300.])
    assert np.all(np.isfinite(ln_f_over_b))


def test_get_event_ln_likelihood_ratio_svd_endtime_mass(coinc_bytes):
    likelihood, gamma1, end_time, mass = \
        p_astro_gstlal.get_event_ln_likelihood_ratio_svd_endtime_mass(
            coinc_bytes)
    assert likelihood == pytest.approx(21.65, abs=0.1)
    assert gamma1 == pytest.approx(90)
    assert end_time == pytest.approx(1174052512)
    assert mass == pytest.approx(3.78, abs=0.1)


def test_load_search_results(coinc_bytes):
    """Fake trigger database returns SVD bank number as 28"""
    likelihood, gamma1, end_time, mass = \
        p_astro_gstlal.get_event_ln_likelihood_ratio_svd_endtime_mass(
            coinc_bytes)
    background, zerolag, svd_banks = \
        p_astro_gstlal.load_search_results(end_time, mass, 6)
    assert np.all(svd_banks == 28)
    background, zerolag, svd_banks = \
        p_astro_gstlal.load_search_results(end_time, mass, 6)
    assert np.all(svd_banks == 28)


def test_compute_p_astro(coinc_bytes, ranking_data_bytes):
    """Test to call :meth:`compute_p_astro` on gracedb event G322589.
    m1 = 2.7, m2 = 1.0 solar mass for this event"""
    files = coinc_bytes, ranking_data_bytes
    p_astros = json.loads(p_astro_gstlal.compute_p_astro(files))
    assert p_astros['BNS'] == pytest.approx(1, abs=1e-3)
    assert p_astros['NSBH'] == pytest.approx(0, abs=1e-3)
    assert p_astros['BBH'] == pytest.approx(0, abs=1e-3)
    assert p_astros['Terr'] == pytest.approx(0, abs=1e-3)



"""Data quality and detector characterization tasks.

These tasks are mostly focused on checking interferometer state vectors. By
design, the [LIGO]_ and [Virgo]_ state vectors share the same definitions for
the first 8 fields.

LIGO also has a [DMT]_ DQ vector that provides some additional instrumental
checks.

References
----------
.. [LIGO] https://wiki.ligo.org/Calibration/TDCalibReview
.. [Virgo] https://dcc.ligo.org/G1801125/
.. [DMT] https://wiki.ligo.org/DetChar/DmtDqVector
"""
import glob

from celery.exceptions import Ignore
from celery.utils.log import get_task_logger
from glue.lal import Cache
from gwpy.timeseries import TimeSeries
import numpy as np

from ..celery import app
from . import gracedb

__author__ = 'Geoffrey Mo <geoffrey.mo@ligo.org>'

log = get_task_logger(__name__)


def create_cache(ifo):
    """Find .gwf files and create cache.

    Parameters
    ----------
    ifo : str
        Interferometer name (e.g. ``H1``).

    Returns
    -------
    :class:`glue.lal.Cache`

    Example
    -------
    >>> create_cache('H1')
    [<glue.lal.CacheEntry at 0x7fbae6b71278>,
      <glue.lal.CacheEntry at 0x7fbae6ae5b38>,
      <glue.lal.CacheEntry at 0x7fbae6ae5c50>,
     ...
      <glue.lal.CacheEntry at 0x7fbae6b15080>,
      <glue.lal.CacheEntry at 0x7fbae6b15828>]

    Note that running this example will return an I/O error, since /dev/shm
    gets overwritten every 300 seconds.

    Notes
    -----
    There are two main ways which this function can fail, which need to
    be accounted for in the future. The first is that the directory
    (typically /dev/shm/llhoft) is found, but the files in question
    corresponding to the timestamp are not in place. This can happen if the
    function is late to the game, and hence the data have been deleted from
    memory and are no longer stored in /dev/shm/llhoft. It can also happen if
    through some asynchronous processes, the call is early, and the data files
    have not yet been written to /dev/shm/llhoft. The second way is if
    /dev/shm/llhoft is not found and hence data never shows up.

    In these cases, the desired behaviour will be for the function to wait a
    period of ~5 seconds and try again. If it still returns an I/O error of
    this type, then the function will return a flag and stop trying (this can
    happen by setting a maximum number of retries to 1).

    This is important for if gwcelery is run locally (and not on a cluster),
    where /dev/shm is inaccessible.
    """
    pattern = app.conf['llhoft_glob'].format(detector=ifo)
    filenames = glob.glob(pattern)
    return Cache.from_urls(filenames)


def check_vector(cache, channel, start, end, bitmask, logic_type='all'):
    """Check timeseries of decimals against a bitmask.
    This is inclusive of the start time and exclusive of the end time, i.e.
    [start, ..., end).

    Parameters
    ----------
    cache : :class:`glue.lal.Cache`
        Cache from which to check.
    channel : str
        Channel to look at, e.g. ``H1:DMT-DQ_VECTOR``.
    start, end : int or float
        GPS start and end times desired.
    bitmask : binary integer
        Bitmask which needs to be 1 in order for the timeseries to pass.
        Example: 0b11 means the 0th and 1st bits need to be 1.
    logic_type : str, optional
        Type of logic to apply for vetoing.
        If ``all``, then all samples in the window must pass the bitmask.
        If ``any``, then one or more samples in the window must pass.

    Returns
    -------
    bool
        True if passes, False otherwise.

    Example
    -------
    Using ``cache`` as output from example in :func:`create_cache`

    >>> check_vector(cache, 'H1:DMT-DQ_VECTOR', 1215370032, 1215370034,
                     0b11, 'all')
    True

    Notes
    -----
    For timeseries of under ~300 samples, it is slightly more
    efficient to check each sample in the series instead of checking the
    entire series, as is done here.
    """
    if logic_type not in ('any', 'all'):
        raise ValueError("logic_type must be either 'all' or 'any'.")

    try:
        timeseries = TimeSeries.read(cache, channel, start=start, end=end)
    except IndexError:
        # FIXME: figure out how to get access to low-latency frames outside
        # of the cluster. Until we figure that out, actual I/O errors have
        # to be non-fatal.
        log.exception('Failed to read from low-latency frame files')
        return None

    # FIXME: Explicitly cast to Python bool because ``np.all([1]) is True``
    # does not evaluate to True
    return bool(getattr(np, logic_type)(timeseries.value & bitmask == bitmask))


@app.task(shared=False)
def check_vectors(event, superevent_id, start, end):
    """Perform data quality checks for an event. This includes checking the DQ
    overflow vector (DMT-DQ_VECTOR) for LIGO and the first and second bits of
    the calibration state vectors for LIGO (GDS-CALIB_STATE_VECTOR) and Virgo
    (DQ_ANALYSIS_STATE_VECTOR). These vectors and the bitmasks for each are
    defined in dictionaries in ``celery.py``.

    The results of these checks are logged into the superevent specified
    by ``superevent_id``, and ``DQOK`` or ``DQV`` labels are appended
    as appropriate.

    This skips MDC events.

    Parameters
    ----------
    event : dict
        Details of event.
    superevent_id : str
        GraceID of event to which to log.
    start, end : int or float
        GPS start and end times desired.

    Returns
    -------
    event : dict
        Details of event.
    """
    # Skip MDC events.
    if event.get('search') == 'MDC':
        log.info('Skipping state vector checks because %s is an MDC',
                 event['graceid'])
        return event

    instruments = event['instruments'].split(',')
    pre, post = app.conf['check_vector_prepost'][event['pipeline']]
    start, end = start - pre, end + post

    ifos = {key.split(':')[0] for key in
            app.conf['llhoft_state_vectors'].keys()}
    caches = {ifo: create_cache(ifo) for ifo in ifos}
    states = {key: check_vector(caches[key.split(':')[0]], key, start,
                                end, *value)
              for key, value in app.conf['llhoft_state_vectors'].items()}
    active_states = {key: value for key, value in states.items()
                     if key.split(':')[0] in instruments}

    if None in active_states.values():
        overall_active_state = None
    elif False in active_states.values():
        overall_active_state = False
    else:
        assert all(active_states.values())
        overall_active_state = True

    fmt = """detector state for active instruments is {}. For all instruments,
    channels good ({}), bad ({}), unknown ({})"""
    msg = fmt.format(
        {None: 'unknown', False: 'bad', True: 'good'}[overall_active_state],
        ', '.join(k for k, v in states.items() if v is True),
        ', '.join(k for k, v in states.items() if v is False),
        ', '.join(k for k, v in states.items() if v is None),
    )

    gracedb.client.writeLog(superevent_id, msg, tag_name=['data_quality'])

    if overall_active_state is True:
        gracedb.create_label('DQOK', superevent_id)
    elif overall_active_state is False:
        gracedb.create_label('DQV', superevent_id)
        # Halt further proessing of canvas
        raise Ignore('vetoed by state vector')
    return event



"""Routing of LVAlert messages to other tasks."""
import json
from urllib.error import URLError

from celery import chain, group
from celery.exceptions import Ignore
from ligo.gracedb.rest import HTTPError

from ..celery import app
from . import bayestar
from . import circulars
from .core import identity
from . import detchar
from . import em_bright
from . import gcn
from . import gracedb
from . import ligo_fermi_skymaps
from . import lvalert
from . import raven
from . import skymaps
from . import p_astro_gstlal


@lvalert.handler('superevent',
                 'test_superevent',
                 shared=False)
def handle_superevent(alert):
    """Schedule annotations for new superevents.

    After waiting for a time specified by the
    :obj:`~gwcelery.celery.Base.orchestrator_timeout` configuration variable
    for the choice of preferred event to settle down, this task peforms data
    quality checks with :meth:`gwcelery.tasks.detchar.check_vectors` and sends
    calls :meth:`~gwcelery.tasks.orchestrator.preliminary_alert` to send a
    preliminary GCN notice.
    """

    if alert['alert_type'] != 'new':
        return

    superevent_id = alert['object']['superevent_id']
    start = alert['object']['t_start']
    end = alert['object']['t_end']

    (
        get_preferred_event.si(superevent_id).set(
            countdown=app.conf['orchestrator_timeout']
        )
        |
        gracedb.get_event.s()
        |
        detchar.check_vectors.s(superevent_id, start, end)
        |
        preliminary_alert.s(superevent_id)
    ).apply_async()


@lvalert.handler('cbc_gstlal',
                 'cbc_gstlal-spiir',
                 'cbc_pycbc',
                 'cbc_mbtaonline',
                 'test_gstlal',
                 'test_gstlal-spiir',
                 'test_pycbc',
                 'test_mbtaonline',
                 shared=False)
def handle_cbc_event(alert):
    """Peform annotations for CBC events that depend on pipeline-specific
    matched-filter parameter estimates.

    Notes
    -----

    This LVAlert message handler is triggered by updates that include the files
    ``psd.xml.gz`` and ``ranking_data.xml.gz``. The table below lists which
    files are created as a result, and which tasks generate them.

    ============================== =====================================================
    File                           Task
    ============================== =====================================================
    ``bayestar.fits.gz``           :meth:`gwcelery.tasks.bayestar.localize`
    ``source_classification.json`` :meth:`gwcelery.tasks.em_bright.classifier`
    ``p_astro_gstlal.json``        :meth:`gwcelery.tasks.p_astro_gstlal.compute_p_astro`
    ============================== =====================================================
    """  # noqa: E501

    if alert['alert_type'] != 'update':
        return

    graceid = alert['uid']
    filename = alert.get('file')

    if filename == 'psd.xml.gz':
        (
            group(
                gracedb.download.s('coinc.xml', graceid),
                gracedb.download.s('psd.xml.gz', graceid)
            )
            |
            # FIXME: group(A, B) | group(C, D) does not pass the results from
            # tasks A and B to tasks C and D without this.
            identity.s()
            |
            group(
                bayestar.localize.s(graceid)
                |
                gracedb.upload.s(
                    'bayestar.fits.gz', graceid,
                    'sky localization complete', ['sky_loc', 'lvem']
                )
                |
                gracedb.create_label.si('SKYMAP_READY', graceid),

                em_bright.classifier.s(graceid)
                |
                gracedb.upload.s(
                    'source_classification.json', graceid,
                    'source classification complete', ['em_bright', 'lvem']
                )
                |
                gracedb.create_label.si('EMBRIGHT_READY', graceid)
            )
        ).delay()
    elif filename == 'ranking_data.xml.gz':
        (
            gracedb.get_event.s(graceid)
            |
            continue_if.s(group='CBC', pipeline='gstlal')
            |
            group(
                gracedb.download.si('coinc.xml', graceid),
                gracedb.download.si('ranking_data.xml.gz', graceid)
            )
            |
            p_astro_gstlal.compute_p_astro.s()
            |
            gracedb.upload.s(
                'p_astro_gstlal.json', graceid,
                'p_astro computation complete'
            )
            |
            gracedb.create_label.si('PASTRO_READY', graceid)
        ).delay()


@lvalert.handler('superevent',
                 'external_fermi',
                 'external_fermi_grb',
                 'external_grb',
                 'external_snews',
                 'external_snews_supernova',
                 'external_swift',
                 shared=False)
def handle_superevents_externaltriggers(alert):
    """Parse an LVAlert message related to superevents/external triggers and
    dispatch it to other tasks.

    Notes
    -----

    This LVAlert message handler is triggered by creating a new superevent or
    external trigger event, or applying the ``EM_COINC`` label to any
    superevent:

    * Any new event triggers a coincidence search with
      :meth:`gwcelery.tasks.raven.coincidence_search`.
    * The ``EM_COINC`` label triggers the creation of a combined GW-GRB sky map
      using :meth:`gwcelery.tasks.ligo_fermi_skymaps.create_combined_skymap`.
    """
    # Determine GraceDb ID
    graceid = alert['uid']

    if alert['alert_type'] == 'new':
        raven.coincidence_search(graceid, alert['object']).delay()
    elif alert['alert_type'] == 'label':
        if 'EM_COINC' in alert['description'] and graceid.startswith('S'):
            ligo_fermi_skymaps.create_combined_skymap(graceid).delay()


@app.task(autoretry_for=(HTTPError, URLError, TimeoutError),
          default_retry_delay=20.0, retry_backoff=True,
          retry_kwargs=dict(max_retries=500), shared=False)
def download(*args, **kwargs):
    """Download a file from GraceDb.

    This works just like :func:`gwcelery.tasks.gracedb.download`, except that
    it is retried for both :class:`TimeoutError` and
    :class:`~urllib.error.URLError`. In particular, it will be retried for 404
    (not found) errors."""
    return gracedb.download(*args, **kwargs)


@gracedb.task(shared=False)
def get_preferred_event(superevent_id):
    """Determine preferred event for a superevent by querying GraceDb.

    This works just like :func:`gwcelery.tasks.gracedb.get_superevent`, except
    that it returns only the preferred event, and not the entire GraceDb JSON
    response."""
    return gracedb.get_superevent(superevent_id)['preferred_event']


@app.task(shared=False)
def continue_if(event, **kwargs):
    """Continue processing if an event's properties match given values, else
    halt the rest of the canvas by raising :class:`celery.exceptions.Ignore`.

    Example
    -------
    .. code-block:: python

        (
            gracedb.get_event.s('G28048')
            |
            continue_if.s(group='CBC')
            |
            ...
        ).apply_async()
    """
    if all(event.get(key) == value for key, value in kwargs.items()):
        return event
    else:
        raise Ignore('This is not a {} event.'.format(group))


@gracedb.task(shared=False)
def create_voevent(em_bright_json, *args, **kwargs):
    """Create a VOEvent record from an EM bright JSON file.

    Parameters
    ----------
    em_bright_json : str, bytes, None
        The JSON contents of a source classification file generated by
        :meth:`gwcelery.tasks.em_bright.classifier`, or None
    \*args
        Additional positional arguments passed to
        :meth:`gwcelery.tasks.gracedb.create_voevent`.
    \*\*kwargs
        Additional keyword arguments passed to
        :meth:`gwcelery.tasks.gracedb.create_voevent`.

    Returns
    -------
    str
        The filename of the newly created VOEvent.
    """
    kwargs = dict(kwargs)
    if em_bright_json is not None:
        data = json.loads(em_bright_json)
        kwargs['ProbHasNS'] = 0.01 * data['Prob NS2']
        kwargs['ProbHasRemnant'] = 0.01 * data['Prob EMbright']
    return gracedb.create_voevent(*args, **kwargs)


@app.task(ignore_result=True, shared=False)
def preliminary_alert(event, superevent_id):
    """Produce a preliminary alert by copying any sky maps.

    This consists of the following steps:

    1.   Copy any sky maps and source classification from the preferred event
         to the superevent.
    2.   Create standard annotations for sky maps including all-sky plots by
         calling :meth:`gwcelery.tasks.skymaps.annotate_fits`.
    3.   Create a preliminary VOEvent.
    4.   Send the VOEvent to GCN.
    5.   Apply the GCN_PRELIM_SENT label to the superevent.
    6.   Create and upload a GCN Circular draft.
    """
    preferred_event_id = event['graceid']

    if event['group'] == 'CBC':
        skymap_filename = 'bayestar.fits.gz'
        skymap_image_filename = 'bayestar.png'
        skymap_type = 'BAYESTAR'
    elif event['pipeline'] == 'CWB':
        skymap_filename = 'skyprobcc_cWB.fits'
        skymap_image_filename = 'skyprobcc_cWB.png'
        skymap_type = 'CWB'
    elif event['pipeline'] == 'LIB':
        skymap_filename = 'LIB.fits.gz'
        skymap_image_filename = 'LIB.png'
        skymap_type = 'LIB'
    else:
        skymap_filename = skymap_type = skymap_image_filename = None

    # Start with a blank canvas (literally).
    canvas = chain()

    # If there is a sky map, then copy it to the superevent and create plots.
    if skymap_filename is not None:
        canvas |= (
            download.s(skymap_filename, preferred_event_id)
            |
            group(
                gracedb.upload.s(
                    skymap_filename,
                    superevent_id,
                    message='{} localization copied from {}'.format(
                        skymap_type, preferred_event_id),
                    tags=['sky_loc', 'lvem']
                )
                |
                gracedb.create_label.si('SKYMAP_READY', superevent_id),

                skymaps.annotate_fits(
                    skymap_filename,
                    skymap_filename.partition('.fits')[0],
                    superevent_id,
                    ['sky_loc', 'lvem']
                )
            )
        )

    # If this is a CBC event, then copy the EM bright classification.
    if event['group'] == 'CBC':
        canvas |= (
            download.si('source_classification.json', preferred_event_id)
            |
            gracedb.upload.s(
                'source_classification.json',
                superevent_id,
                message='Source classification copied from {}'.format(
                    preferred_event_id),
                tags=['em_bright', 'lvem']
            )
            |
            gracedb.create_label.si('EMBRIGHT_READY', superevent_id)
            |
            download.si('source_classification.json', superevent_id)
        )
    else:
        canvas |= identity.si(None)

    # Send GCN notice and upload GCN circular draft.
    canvas |= (
        create_voevent.s(
            superevent_id, 'preliminary',
            skymap_type=skymap_type,
            skymap_filename=skymap_filename,
            skymap_image_filename=skymap_image_filename
        )
        |
        group(
            gracedb.download.s(superevent_id)
            |
            gcn.send.s()
            |
            gracedb.create_label.si('GCN_PRELIM_SENT', superevent_id),

            circulars.create_circular.si(superevent_id)
            |
            gracedb.upload.s(
                'circular.txt',
                superevent_id,
                'Automated circular'
            )
        )
    )

    canvas.apply_async()



"""Provide subcommand to manage GWCelery through HTCondor."""
import os
import shlex
import subprocess
import sys

from celery.bin.base import Command
import lxml.etree
import pkg_resources

SUBMIT_FILE = pkg_resources.resource_filename(__name__, 'data/gwcelery.sub')
CONSTRAINTS = ('-constraint', 'JobBatchName=="gwcelery"')


def run_exec(*args):
    print(' '.join(shlex.quote(arg) for arg in args))
    os.execvp(args[0], args)


def submit():
    """Submit all GWCelery jobs to HTCondor (if not already running)."""
    status = subprocess.check_output(('condor_q', '-xml') + CONSTRAINTS)
    classads = lxml.etree.fromstring(status)
    if classads.find('.//c') is None:
        run_exec('condor_submit', SUBMIT_FILE)
    else:
        print('error: GWCelery jobs are already running.\n'
              'You must first remove exist jobs with "gwcelery condor rm".\n'
              'To see the status of those jobs, run "gwcelery condor q".',
              file=sys.stderr)
        sys.exit(1)


def rm():
    """Remove all GWCelery jobs."""
    run_exec('condor_rm', *CONSTRAINTS)


def hold():
    """Put all GWCelery jobs on hold."""
    run_exec('condor_hold', *CONSTRAINTS)


def release():
    """Release all GWCelery jobs from hold status."""
    run_exec('condor_release', *CONSTRAINTS)


def q():
    """Show status of all GWCelery jobs."""
    run_exec('condor_q', '-nobatch', *CONSTRAINTS)


class CondorCommand(Command):
    """Shortcuts for HTCondor commands to manage deployment of GWCelery on LIGO
    Data Grid clusters."""

    def add_arguments(self, parser):
        subparsers = parser.add_subparsers()
        for func in [submit, rm, hold, release, q]:
            subparser = subparsers.add_parser(func.__name__, help=func.__doc__)
            subparser.set_defaults(func=func)

    def run(self, func=None, **kwargs):
        func()



from unittest.mock import patch
from xml.sax import SAXParseException

from astropy import table
from astropy.io import fits
import numpy as np
import pkg_resources
import pytest

from ..tasks.bayestar import localize
from ..util.tempfile import NamedTemporaryFile


def test_localize_bad_psd():
    """Test running BAYESTAR with a pad PSD file"""
    # Test data
    coinc = pkg_resources.resource_string(__name__, 'data/coinc.xml')
    psd = b''

    # Run function under test
    with pytest.raises(SAXParseException):
        localize((coinc, psd), 'G211117')


def mock_bayestar(*args, **kwargs):
    return table.Table({'UNIQ': np.arange(4, 16, dtype=np.uint64),
                        'PROBDENSITY': np.ones(12),
                        'DISTMU': np.ones(12),
                        'DISTSIGMA': np.ones(12),
                        'DISTNORM': np.ones(12)})


@pytest.mark.parametrize('disabled_detectors', [None,
                                                ['H1', 'L1'],
                                                ['H1', 'L1', 'V1']])
@patch('ligo.skymap.bayestar.localize', mock_bayestar)
def test_localize(disabled_detectors):
    """Test running BAYESTAR on G211117"""
    # Test data
    coinc = pkg_resources.resource_string(__name__, 'data/coinc.xml')
    psd = pkg_resources.resource_string(__name__, 'data/psd.xml.gz')

    # Run function under test
    fitscontent = localize(
        (coinc, psd), 'G211117', disabled_detectors=disabled_detectors)

    with NamedTemporaryFile(content=fitscontent) as fitsfile:
        url = fits.getval(fitsfile.name, 'REFERENC', 1)
        assert url == 'https://gracedb.invalid/events/G211117'



import logging

from pkg_resources import resource_filename
import pytest

from ..celery import app
from ..tasks import detchar


@pytest.fixture
def llhoft_glob_pass():
    old = app.conf['llhoft_glob']
    app.conf['llhoft_glob'] = resource_filename(
        __name__, 'data/llhoft/pass/{detector}/*.gwf')
    yield
    app.conf['llhoft_glob'] = old


@pytest.fixture
def llhoft_glob_fail():
    old = app.conf['llhoft_glob']
    app.conf['llhoft_glob'] = resource_filename(
        __name__, 'data/llhoft/fail/{detector}/*.gwf')
    yield
    app.conf['llhoft_glob'] = old


def test_create_cache(llhoft_glob_fail):
    assert len(detchar.create_cache('L1')) == 1


def test_check_vector(llhoft_glob_pass):
    channel = 'H1:DMT-DQ_VECTOR'
    start = 1214714160
    end = 1214714164
    cache = detchar.create_cache('H1')
    assert detchar.check_vector(cache, channel, start, end, 0b11, 'any')
    assert not detchar.check_vector(cache, channel, start, end, 0b1111, 'any')


def test_check_vectors_skips_mdc(caplog):
    """Test that state vector checks are skipped for MDC events."""
    caplog.set_level(logging.INFO)
    detchar.check_vectors({'search': 'MDC', 'graceid': 'M1234'}, 'S123', 0, 1)
    messages = [record.message for record in caplog.records]
    assert 'Skipping state vector checks because M1234 is an MDC' in messages



from unittest.mock import patch
from pkg_resources import resource_string

from ligo.gracedb import rest

from ..tasks import gracedb


def test_create_event(monkeypatch):

    class MockResponse(object):

        def json(self):
            return {'graceid': 'T12345'}

    class MockGraceDb(object):

        def createEvent(self, group, pipeline, filename,  # noqa: N802
                        search=None, labels=None, offline=False,
                        filecontents=None, **kwargs):
            assert group == 'group'
            assert pipeline == 'pipeline'
            assert filename == 'initial.data'
            assert search == 'search'
            assert filecontents == 'filecontents'
            return MockResponse()

    monkeypatch.setattr('gwcelery.tasks.gracedb.client', MockGraceDb())

    graceid = gracedb.create_event('filecontents', 'search', 'pipeline',
                                   'group')
    assert graceid == 'T12345'


@patch('gwcelery.tasks.gracedb.client', autospec=rest.GraceDb)
def test_create_label(mock_gracedb):
    # Run function under test.
    gracedb.create_label('label', 'graceid')

    # Check that one file was downloaded.
    mock_gracedb.writeLabel.assert_called_once_with('graceid', 'label')


@patch('gwcelery.tasks.gracedb.client', autospec=rest.GraceDb)
def test_create_tag(mock_gracedb):
    # Run function under test.
    gracedb.create_tag('tag', 'n', 'graceid')

    # Check that one file was downloaded.
    mock_gracedb.createTag.assert_called_once_with('graceid', 'n', 'tag')


@patch('gwcelery.tasks.gracedb.client', autospec=rest.GraceDb)
def test_create_voevent(mock_gracedb):
    # Run function under test.
    gracedb.create_voevent('graceid', 'voevent_type',
                           skymap_filename='skymap_filename',
                           skymap_type='skymap_type')

    # Check that one file was downloaded.
    mock_gracedb.createVOEvent.assert_called_once_with(
        'graceid', 'voevent_type',
        skymap_filename='skymap_filename',
        skymap_type='skymap_type')


@patch('gwcelery.tasks.gracedb.client', autospec=rest.GraceDb)
def test_download(mock_gracedb):
    # Run function under test.
    gracedb.download('filename', 'graceid')

    # Check that one file was downloaded.
    mock_gracedb.files.assert_called_once_with('graceid', 'filename', raw=True)


def test_get_log(monkeypatch):

    class logs(object):  # noqa: N801

        def json(self):
            return {'log': 'stuff'}

    class MockGraceDb(object):

        def logs(self, graceid):
            assert graceid == 'graceid'
            return logs()

    monkeypatch.setattr('gwcelery.tasks.gracedb.client', MockGraceDb())

    # Run function under test.
    ret = gracedb.get_log('graceid')

    assert ret == 'stuff'


@patch('gwcelery.tasks.gracedb.client', autospec=rest.GraceDb)
def test_get_superevent(mock_gracedb):
    # Run function under test.
    gracedb.get_superevent('graceid')

    # Check that one file was downloaded.
    mock_gracedb.superevent.assert_called_once_with('graceid')


@patch('gwcelery.tasks.gracedb.client', autospec=rest.GraceDb)
def test_upload(mock_gracedb):
    # Run function under test.
    gracedb.upload('filecontents', 'filename', 'graceid', 'message', 'tags')

    # Check that one file was uploaded.
    mock_gracedb.writeLog.assert_called_once_with(
        'graceid', 'message', 'filename', 'filecontents', 'tags')


@patch('gwcelery.tasks.gracedb.client', autospec=rest.GraceDb)
def test_get_event(mock_gracedb):
    gracedb.get_event('G123456')
    mock_gracedb.event.assert_called_once_with('G123456')


@patch('gwcelery.tasks.gracedb.client', autospec=rest.GraceDb)
def test_get_events(mock_gracedb):
    gracedb.get_events(query='Some query', orderby=None,
                       count=None, columns=None)
    mock_gracedb.events.assert_called_once_with(query='Some query',
                                                orderby=None, count=None,
                                                columns=None)


@patch('gwcelery.tasks.gracedb.client', autospec=rest.GraceDb)
def test_replace_event(mock_gracedb):
    text = resource_string(__name__, 'data/fermi_grb_gcn.xml')
    gracedb.replace_event(graceid='G123456', payload=text)
    mock_gracedb.replaceEvent.assert_called_once_with(graceid='G123456',
                                                      filename='initial.data',
                                                      filecontents=text)



"""LVAlert client."""
import json
import time

from celery_eternal import EternalTask
from celery.utils.log import get_task_logger
import sleek_lvalert

from ..celery import app
from .core import DispatchHandler
from . import gracedb

log = get_task_logger(__name__)


class _LVAlertDispatchHandler(DispatchHandler):

    def process_args(self, node, payload):
        # Determine GraceDB service URL
        alert = json.loads(payload)
        try:
            try:
                self_link = alert['object']['links']['self']
            except KeyError:
                self_link = alert['object']['self']
        except KeyError:
            log.exception('LVAlert message does not contain an API URL: %r',
                          alert)
            return None, None, None
        base, api, _ = self_link.partition('/api/')
        service = base + api

        if service != gracedb.client.service_url:
            log.warn('ignoring LVAlert message because it is intended for '
                     'GraceDb server %s, but we are set up for server %s',
                     service, gracedb.client.service_url)
            return None, None, None

        return super().process_args(node, alert)


handler = _LVAlertDispatchHandler()
"""Function decorator to register a handler callback for specified LVAlert
message types. The decorated function is turned into a Celery task, which will
be automatically called whenever a matching LVAlert message is received.

Parameters
----------
\*keys
    List of LVAlert message types to accept
\*\*kwargs
    Additional keyword arguments for :meth:`celery.Celery.task`.

Examples
--------
Declare a new handler like this::

    @lvalert.handler('cbc_gstlal',
                     'cbc_gstlal-spiir',
                     'cbc_pycbc',
                     'cbc_mbtaonline')
    def handle_cbc(alert_content):
        # do work here...
"""


@app.task(base=EternalTask, bind=True, shared=False)
def listen(self):
    """Listen for LVAlert messages forever. LVAlert messages are dispatched
    asynchronously to tasks that have been registered with
    :meth:`gwcelery.tasks.lvalert.handler`."""

    log.info('Starting client')
    client = sleek_lvalert.LVAlertClient(server=app.conf['lvalert_host'])
    client.connect()
    client.process(block=False)

    log.info('Updating subscriptions')
    current_subscriptions = set(client.get_subscriptions())
    needed_subscriptions = set(handler.keys())
    client.subscribe(*(needed_subscriptions - current_subscriptions))

    log.info('Listening for pubsub messages')
    client.listen(handler.dispatch)

    while not self.is_aborted():
        time.sleep(1)
    log.info('Disconnecting')
    client.abort()
    log.info('Exiting')



"""Miscellaneous utilities that are useful inside many different tasks."""
from . import proxy
from . import tempfile
from .proxy import *  # noqa
from .tempfile import *  # noqa

__all__ = proxy.__all__ + tempfile.__all__



import sys
from setuptools import setup

setup_requires = ['setuptools >= 30.3.0']
if {'pytest', 'test', 'ptr'}.intersection(sys.argv):
    setup_requires.append('pytest-runner')
if {'build_sphinx'}.intersection(sys.argv):
    setup_requires.extend(['celery_eternal>=0.0.2', 'recommonmark', 'sphinx',
                           'sphinx_celery'])

setup(setup_requires=setup_requires)



import json
import logging
import socket
import struct
from threading import Thread
from time import sleep
from unittest.mock import MagicMock, patch

from gcn.voeventclient import _recv_packet
import lxml.etree
import pkg_resources
import pytest

from ..tasks import gcn
from .. import app

logging.basicConfig(level=logging.INFO)

# Test data
with pkg_resources.resource_stream(
        __name__, 'data/lvalert_voevent.json') as f:
    lvalert = json.load(f)
voevent = lvalert['object']['text']


@pytest.fixture
def broker_thread(monkeypatch):
    queue = [b'foo', b'bar', b'bat']

    def lindex(key, index):
        assert key == gcn._queue_name
        try:
            return queue[index]
        except IndexError:
            return None

    def lpop(key):
        assert key == gcn._queue_name
        try:
            return queue.pop(0)
        except IndexError:
            return None

    # Decrease keepalive time so that we see keepalive packets frequently
    monkeypatch.setattr('gwcelery.tasks.gcn.KEEPALIVE_TIME', 1)
    monkeypatch.setattr('gwcelery.tasks.gcn.broker.backend.client.lindex',
                        lindex, raising=False)
    monkeypatch.setattr('gwcelery.tasks.gcn.broker.backend.client.lpop',
                        lpop, raising=False)

    monkeypatch.setattr('gwcelery.tasks.gcn.broker.is_aborted', lambda: False)
    thread = Thread(target=gcn.broker)
    thread.start()
    sleep(0.1)
    yield
    monkeypatch.setattr('gwcelery.tasks.gcn.broker.is_aborted', lambda: True)
    thread.join()


@pytest.fixture
def wrong_remote_address():
    old = app.conf['gcn_broker_accept_addresses']
    app.conf['gcn_broker_accept_addresses'] = ['192.0.2.0']
    yield
    app.conf['gcn_broker_accept_addresses'] = old


@pytest.fixture
def connection_to_broker(broker_thread):
    with socket.socket() as sock:
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_LINGER,
                        struct.pack('ii', 1, 0))
        sock.connect(('127.0.0.1', 53410))
        yield sock


@pytest.mark.enable_socket
def test_broker_aborted_before_accept(broker_thread, monkeypatch):
    """Test aborting the broker while it is still waiting for connections."""
    monkeypatch.setattr('gwcelery.tasks.gcn.broker.is_aborted', lambda: True)


@pytest.mark.enable_socket
def test_broker_aborted_after_accept(connection_to_broker, monkeypatch):
    """Test aborting the broker after it has accepted a connection."""
    monkeypatch.setattr('gwcelery.tasks.gcn.broker.is_aborted', lambda: True)


@pytest.mark.enable_socket
def test_broker_disconnected(connection_to_broker):
    """Test connection from a broker to a client that immediately closes the
    socket."""
    pass


@pytest.mark.enable_socket
def test_broker_wrong_address(capsys, wrong_remote_address,
                              connection_to_broker):
    """Test that the broker refuses connections from the wrong IP address."""
    assert connection_to_broker.recv(1) == b''


@pytest.mark.enable_socket
def test_broker(connection_to_broker, broker_thread):
    """Test receiving packets from the broker."""
    connection_to_broker.settimeout(2.0)
    packets = [_recv_packet(connection_to_broker) for _ in range(4)]
    assert packets[:3] == [b'foo', b'bar', b'bat']
    assert b'iamalive' in packets[3]


def test_send(monkeypatch):
    mock_rpush = MagicMock()
    monkeypatch.setattr('gwcelery.tasks.gcn.broker.backend.client.rpush',
                        mock_rpush, raising=False)
    gcn.send(b'foo')
    mock_rpush.assert_called_once_with(gcn._queue_name, b'foo')


@pytest.mark.enable_socket
def test_listen(monkeypatch):
    """Test that the listen task would correctly launch gcn.listen()."""
    mock_gcn_listen = MagicMock()
    monkeypatch.setattr('gcn.listen', mock_gcn_listen)
    gcn.listen.run()
    mock_gcn_listen.assert_called_once()


def fake_gcn(notice_type):
    # Check the real GCN notice, which is valid.
    payload = pkg_resources.resource_string(
        __name__, 'data/G298048-1-Initial.gcn.xml')
    root = lxml.etree.fromstring(payload)
    notice_type = str(int(notice_type))
    root.find(".//Param[@name='Packet_Type']").attrib['value'] = notice_type
    return lxml.etree.tostring(root), root


def test_unrecognized_notice_type(caplog):
    """Test handling an unrecognized (enum not defined) notice type."""
    caplog.set_level(logging.WARNING)
    gcn.handler.dispatch(*fake_gcn(10000))
    record, = caplog.records
    assert record.message == 'ignoring unrecognized key: 10000'


def test_unregistered_notice_type(caplog):
    """Test handling an unregistered notice type."""
    caplog.set_level(logging.WARNING)
    gcn.handler.dispatch(*fake_gcn(gcn.NoticeType.SWIFT_UVOT_POS_NACK))
    record, = caplog.records
    assert record.message == ('ignoring unrecognized key: '
                              '<NoticeType.SWIFT_UVOT_POS_NACK: 89>')


@pytest.fixture
def reset_handlers():
    old_handler = dict(gcn.handler)
    gcn.handler.clear()
    yield
    gcn.handler.update(old_handler)


def test_registered_notice_type(reset_handlers):
    @gcn.handler(gcn.NoticeType.AGILE_POINTDIR, gcn.NoticeType.AGILE_TRANS)
    def agile_handler(payload):
        pass

    with patch.object(agile_handler, 'run') as mock_run:
        gcn.handler.dispatch(*fake_gcn(gcn.NoticeType.SWIFT_UVOT_POS_NACK))
        mock_run.assert_not_called()
        gcn.handler.dispatch(*fake_gcn(gcn.NoticeType.AGILE_POINTDIR))
        mock_run.assert_called_once()



"""Celery application configuration."""

from celery import Celery


class Base:
    """Base application configuration."""

    # Celery application settings.
    # Use pickle serializer, because it supports byte values.

    accept_content = ['json', 'pickle']
    event_serializer = 'json'
    result_serializer = 'pickle'
    task_serializer = 'pickle'

    # GWCelery-specific settings.

    gcn_broker_address = ':5341'
    """The VOEvent broker will bind to this address to send GCNs.
    This should be a string of the form `host:port`. If `host` is empty,
    then listen on all available interfaces."""

    gcn_broker_accept_addresses = ['capella2.gsfc.nasa.gov']
    """List of hosts from which the broker will accept connections."""

    gcn_client_address = '68.169.57.253:8096'
    """The VOEvent listener will connect to this address to receive GCNs.

    We are temporarily using the pre-registered port 8096 for receiving
    proprietary LIGO/Virgo alerts on emfollow.ligo.caltech.edu. This means that
    the capability to receive GCNs requires setting up a site configuration in
    advance with Scott Barthelmey.

    Once we switch to sending public alerts exclusively, then we can switch
    back to using port 8099 for anonymous access, requiring no prior site
    configuration."""

    superevent_d_t_start = {'gstlal': 1.0,
                            'gstlal-spiir': 1.0,
                            'pycbc': 1.0,
                            'mbtaonline': 1.0}
    """Pipeline based lower extent of superevent segments.
    For cwb and lib this is decided from extra attributes."""

    superevent_d_t_end = {'gstlal': 1.0,
                          'gstal-spiir': 1.0,
                          'pycbc': 1.0,
                          'mbtaonline': 1.0}
    """Pipeline based upper extent of superevent segments
    For cwb and lib this is decided from extra attributes."""

    superevent_query_d_t_start = 100.
    """Lower extent of superevents query"""

    superevent_query_d_t_end = 100.
    """Upper extent of superevents query"""

    superevent_default_d_t_start = 1.0
    """Default lower extent of superevent segments"""

    superevent_default_d_t_end = 1.0
    """Default upper extent for superevent segments"""

    superevent_far_threshold = 1 / 3600
    """Maximum false alarm rate to consider events superevents."""

    orchestrator_timeout = 15.0
    """The orchestrator will wait this many seconds from the time of the
    creation of a new superevent to the time that annotations begin, in order
    to let the superevent manager's decision on the preferred event
    stabilize."""

    check_vector_prepost = {'gstlal': [2, 2],
                            'gstlal-spiir': [2, 2],
                            'pycbc': [2, 2],
                            'MBTAOnline': [2, 2],
                            'LIB': [0.5, 0.5],
                            'CWB': [0.5, 0.5],
                            'HardwareInjection': [2, 2],
                            'Swift': [2, 2],
                            'Fermi': [2, 2],
                            'SNEWS': [2, 2]}
    """Seconds before and after the superevent start and end times which the DQ
    vector check will include in its check. Pipeline dependent."""

    llhoft_glob = '/dev/shm/llhoft/{detector}_O2/*.gwf'
    """File glob for low-latency h(t) frames."""

    llhoft_state_vectors = {'H1:DMT-DQ_VECTOR': (0b11, 'all'),
                            'L1:DMT-DQ_VECTOR': (0b11, 'all'),
                            'H1:GDS-CALIB_STATE_VECTOR': (0b11, 'all'),
                            'L1:GDS-CALIB_STATE_VECTOR': (0b11, 'all'),
                            'V1:DQ_ANALYSIS_STATE_VECTOR': (0b11, 'all')}
    """Low-latency h(t) state vector configuration. This is a dictionary
    mapping channel names (e.g. ``H1:DMT-DQ_VECTOR``) to tuples consisting of a
    bitmask to be logically ANDed with the state vector, and the string ``any``
    or ``all`` to indicate how many samples in the segment must satisfy the
    bitmask."""

    p_astro_gstlal_trigger_db = '/home/gstlalcbc/observing/3/online/trigs/H1L1-ALL_LLOID-0-2000000000.sqlite'  # noqa: E501
    """Gstlal trigger database location in CIT"""

    p_astro_gstlal_ln_likelihood_threshold = 6
    """log likelihood threshold"""

    p_astro_gstlal_prior_type = "Uniform"
    """Prior type to be used. Options: Uniform, Jeffreys"""


class Production(Base):
    """Application configuration for ``gracedb.ligo.org``."""

    lvalert_host = 'lvalert.cgca.uwm.edu'
    """LVAlert host."""

    gracedb_host = 'gracedb.ligo.org'
    """GraceDb host."""


class Test(Base):
    """Application configuration for ``gracedb-test.ligo.org``."""

    lvalert_host = 'lvalert-test.cgca.uwm.edu'
    """LVAlert host."""

    gracedb_host = 'gracedb-test.ligo.org'
    """GraceDb host."""


class Development(Test):
    """Application configuration for ``gracedb-dev1.ligo.org``."""

    gracedb_host = 'gracedb-dev1.ligo.org'
    """GraceDb host."""


class Playground(Test):
    """Application configuration for ``gracedb-playground.ligo.org``."""

    gracedb_host = 'gracedb-playground.ligo.org'
    """GraceDb host."""


# Celery application object.
# Use redis broker, because it supports locks (and thus singleton tasks).
app = Celery('gwcelery', broker='redis://', config_source=Playground)

# Use the same URL for both the result backend and the broker.
app.conf['result_backend'] = app.conf.broker_url



import json

import lxml.etree
import pkg_resources
import pytest

from ..tasks.gcn.validate import validate_voevent

# Test data
with pkg_resources.resource_stream(
        __name__, 'data/lvalert_voevent.json') as f:
    lvalert = json.load(f)
voevent = lvalert['object']['text']


@pytest.fixture
def fake_gcn(celeryconf, monkeypatch):

    def mock_download(filename, graceid):
        assert filename == 'G298048-1-Initial.xml'
        assert graceid == 'G298048'
        return pkg_resources.resource_string(__name__, 'data/' + filename)

    def mock_get_log(graceid):
        assert graceid == 'G298048'
        return json.loads(
            pkg_resources.resource_string(__name__, 'data/G298048_log.json'))

    def mock_create_tag(tag, n, graceid):
        assert tag == 'gcn_received'
        assert n == 532
        assert graceid == 'G298048'

    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.download', mock_download)
    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.get_log', mock_get_log)
    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.create_tag', mock_create_tag)

    # Check the real GCN notice, which is valid.
    payload = pkg_resources.resource_string(
        __name__, 'data/G298048-1-Initial.gcn.xml')
    root = lxml.etree.fromstring(payload)
    yield root


def test_validate_voevent(fake_gcn):
    """Test that the fake GCN notice matches what we actually sent."""
    validate_voevent(lxml.etree.tostring(fake_gcn))


def test_validate_voevent_mismatched_param(fake_gcn):
    """Test that we correctly detect mismatched parameter values in GCNs."""
    xpath = ".//Param[@name='SKYMAP_URL_FITS_BASIC']"
    fake_gcn.find(xpath).attrib['value'] = ('wrong, you fool!')
    with pytest.raises(
            AssertionError,
            match="^GCN parameter 'SKYMAP_URL_FITS_BASIC' has value 'wrong, "
            "you fool!', but original VOEvent parameter 'skymap_fits_basic' "
            "has value 'https://gracedb.ligo.org/apibasic/events/G298048/"
            "files/bayestar.fits.gz'$"):
        validate_voevent(lxml.etree.tostring(fake_gcn))


def test_validate_voevent_unexpected_param(fake_gcn, monkeypatch):
    """Test that we correctly detect unexpected paremeters in GCNs."""

    def mock_download(filename, graceid):
        assert filename == 'G298048-1-Initial.xml'
        assert graceid == 'G298048'
        payload = pkg_resources.resource_string(__name__, 'data/' + filename)
        root = lxml.etree.fromstring(payload)
        elem = root.find(".//Group[@type='GW_SKYMAP']")
        elem.getparent().remove(elem)
        return lxml.etree.tostring(root)

    monkeypatch.setattr('gwcelery.tasks.gracedb.download', mock_download)

    with pytest.raises(
            AssertionError,
            match="^GCN has unexpected parameter: 'SKYMAP_URL_FITS_[A-Z]*'$"):
        validate_voevent(lxml.etree.tostring(fake_gcn))


def test_validate_voevent_missing_param(fake_gcn, monkeypatch):
    """Test that we correctly detect unexpected paremeters in GCNs."""
    elem = fake_gcn.find(".//Param[@name='SKYMAP_URL_FITS_BASIC']")
    elem.getparent().remove(elem)
    with pytest.raises(
            AssertionError,
            match="^GCN is missing parameter: 'SKYMAP_URL_FITS_BASIC'$"):
        validate_voevent(lxml.etree.tostring(fake_gcn))


def test_validate_voevent_wrong_ivorn_path(fake_gcn):
    """Test that we correctly detect the wrong IVORN path."""
    fake_gcn.attrib['ivorn'] = 'ivo://nasa.gsfc.gcn/wrong#G298048-1-Initial'
    with pytest.raises(
            AssertionError, match="^IVORN has unexpected path: '/wrong'$"):
        validate_voevent(lxml.etree.tostring(fake_gcn))


def test_validate_voevent_wrong_ivorn_netloc(fake_gcn):
    """Test that we correctly detect the wrong IVORN netloc."""
    fake_gcn.attrib['ivorn'] = 'ivo://wrong/LVC#G298048-1-Initial'
    with pytest.raises(
            AssertionError, match="^IVORN has unexpected netloc: 'wrong'$"):
        validate_voevent(lxml.etree.tostring(fake_gcn))


def test_validate_voevent_wrong_ivorn_scheme(fake_gcn):
    """Test that we correctly detect the wrong IVORN scheme."""
    fake_gcn.attrib['ivorn'] = 'wrong://nasa.gsfc.gcn/LVC#G298048-1-Initial'
    with pytest.raises(
            AssertionError, match="^IVORN has unexpected scheme: 'wrong'$"):
        validate_voevent(lxml.etree.tostring(fake_gcn))



"""
Sphinx gets confused by :class:`celery.local.PromiseProxy` objects.
If running under Sphinx, substitute a dummy class for
:class:`~celery.local.PromiseProxy`.
"""
import os
import sys

__all__ = ('PromiseProxy',)

prog = os.path.basename(sys.argv[0])
if prog != 'sphinx-build' and 'build_sphinx' not in sys.argv:
    from celery.local import PromiseProxy
else:
    class PromiseProxy:

        def __init__(self, *args, **kwargs):
            pass
